#define _GNU_SOURCE
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

#include "funcs.h"

#ifdef DEBUG
#define CORRUPT_CHECK
#define POISON_FREES
#endif

#undef DEBUG_MEMORY_USE

/* Globals */

/* Functions */

/*---------------------------------------------------------------------------*
 *
 * Memory allocation functions.
 *
 *---------------------------------------------------------------------------*/

/*
 * realloc() with error checking.
 */
void *safe_realloc(void *ptr,size_t size){
	void *p = realloc(ptr,size);

	if (size && (p == NULL)){
		perror("realloc()");
		exit(1);
	}
	return p;
}

/*
 * malloc() with error checking.
 */
void *safe_malloc(size_t size){
	void *p = malloc(size);

	if (p == NULL){
		perror("malloc()");
		exit(1);
	}
	return p;
}

/*
 * strdup() with error checking.
 */
char *safe_strdup(char *s){
	char *p = strdup(s);

	if (p == NULL){
		perror("strdup()");
		exit(1);
	}
	return p;
}

/*
 * strdup() with variable slack and error checking.
 */
char *safe_strdup_with_slack(char *s,size_t slack){
	char *p = malloc(strlen(s) + 1 + slack);

	if (p == NULL){
		perror("malloc()");
		exit(1);
	}
	strcpy(p,s);
	return p;
}

/*---------------------------------------------------------------------------*
 *
 * Text/string manipulation functions.
 *
 *---------------------------------------------------------------------------*/

/*
 * Change a string to lowercase.
 */
unsigned char *string_tolower(unsigned char *s){
	unsigned char *b = s;

	for (;*s;s++)
		*s = tolower(*s);
	return b;
}

/*
 * Count occurrences of a character in a string.
 */
int string_char_occurrences(char *s,char c){
	int count = 0;
	for (;*s;s++)
		if (*s == c)
			count++;
	return count;
}

/*
 * Eat newlines off of the end until there are no more.
 */
char *chomp(char *s){
	size_t l = strlen(s);

	while (l-- > 0){
		if ((s[l] != '\r') &&
		    (s[l] != '\n'))  
			break;
	}
	s[l + 1] = '\0';
	return s;
}

/*
 * Skip ahead to non-whitespace.
 */
char *skip_whitespace(char *s){
	while ((*s == ' ') || (*s == '\t'))
		s++;
	return s;
}

/*
 * Skip ahead to whitespace.
 */
char *skip_non_whitespace(char *s){
	while ((*s) && (*s != ' ') && (*s != '\t'))
		s++;
	return s;
}

/*
 * Skip past whitespace.
 */
char *skip_past_whitespace(char *s){
	return skip_whitespace(skip_non_whitespace(s));
}

/*
 * Skip past comma, with whitespace hacks.
 */
char *skip_past_comma(char *s){
	while ((*s) && (*s != ',') && (*s != ' ') && (*s != '\t'))
		s++;
	if (*s == ',')
		s++;
	return s;
}

/*
 * Skip past slash.
 */
char *skip_past_char(char *s,char c){
	while ((*s) && (*s != c))
		s++;
	if (*s == c)
		s++;
	return s;
}

/*
 * Eat whitespace on either end of string.
 */
char *eat_whitespace(char *s){
	char *e,*d;
	while ((*s == ' ') || (*s == '\t'))
		s++;
	d = s;
	for (;;){
		e = skip_non_whitespace(d);
		d = skip_whitespace(e);
		if (!*d)
			return e;
	}
}

/*
 * Skip numeric characters.
 */
char *skip_numeric(char *s){
	while ((*s >= '0') && (*s <= '9'))
		s++;
	return s;
}

/*
 * Skip non-numeric characters.
 */
char *skip_non_numeric(char *s){
	while ((*s) && ((*s < '0') || (*s > '9')))
		s++;
	return s;
}

/*
 * Split chunk at whitespace into array of arguments.
 */
void split_args(unsigned char *c,int len,struct args *args){
	int arg_size = 16;
	int i = 0;
	int ofs = 0;
	char *s;

	args->argbuf = s = safe_strdup(c);
	args->arg = args->argc = (unsigned char **)NULL;

	/*
	 * c = pointer to continuous words (original string).
	 * s = pointer to split words (duped string, argbuf).
	 */

	while (ofs < len){
		for (;ofs < len;ofs++)
			if ((s[ofs] != ' ') &&
			    (s[ofs] != '\t'))
				break;
		if (ofs >= len)
			break;

		if ((!args->arg) || (i + 1 >= arg_size)){
			if (i + 1 >= arg_size)
				arg_size<<= 1;
			args->arg = (unsigned char **)safe_realloc(args->arg,arg_size * sizeof(args->arg));
			args->argc = (unsigned char **)safe_realloc(args->argc,arg_size * sizeof(args->argc));
		}
		args->arg[i] = s + ofs;
		args->argc[i] = c + ofs;
		i++;

		for (;ofs < len;ofs++)
			if ((s[ofs] == ' ') ||
			    (s[ofs] == '\t')){
			    	s[ofs++] = '\0';
				break;
			}
	}
	s[ofs] = '\0';	/* Terminate last argument */
	if (args->arg)
		args->arg[i] = args->argc[i] = (unsigned char *)NULL;
	args->count = i;
}

/* Misc. functions */

int is_hex_digit(char c){
	if (((c >= '0') &&
	     (c <= '9')) ||
	    ((c >= 'a') &&
	     (c <= 'z')) ||
	    ((c >= 'A') &&
	     (c <= 'Z')))
		return 1;
	return 0;
}

int my_atoi(char *s){
	int num;
	char *e = NULL;

	num = (int)strtol(s,&e,10);
	if ((e) && (e != s) && (*e == '\0'))
		return num;
	return -1;
}

int my_hex_atoi(char *s){
	int num;
	char *e = NULL;

	num = (int)strtol(s,&e,16);
	if ((e) && (e != s) && (*e == '\0'))
		return num;
	return -1;
}

/*---------------------------------------------------------------------------*
 *
 * Variable buffer functions.
 *
 *---------------------------------------------------------------------------*/

/*
 * Simple dynamic buffer/string manipulation functions, currently
 * supporting appending and header skipping.
 *
 * Strings and buffer/length style buffers are interchangeable, both
 * when appending and requesting the buffer.  Obviously, strings may
 * terminate prematurely if the buffer was originally appended as a
 * buffer/length style buffer containing one or more '\0' characters.
 */

/*
 * Initialize variable buffer structure values.
 *
 * Call this only on initialization.  Use variable_buffer_clear
 * after initialization -- it will free any existing buffer.
 */
void variable_buffer_init(struct varbuf *vbp){
	vbp->buffer = NULL;
        vbp->buffer_length = vbp->allocated_length =
        vbp->header_skip_length = vbp->last_append_offset = 0;
}

/*
 * Initialize variable buffer structure or free/clear existing buffer.
 */
void variable_buffer_clear(struct varbuf *vbp)
{
	free(vbp->buffer);
	variable_buffer_init(vbp);
}

/*
 * Increase buffer size intelligently, if necessary.
 */
void variable_buffer_size_extend(struct varbuf *vbp,size_t length){
	if (length <= vbp->allocated_length)
		return;

	if (!vbp->allocated_length)
		vbp->allocated_length = INITIAL_VARIABLE_BUFFER_SIZE;

	while (length > vbp->allocated_length)
		vbp->allocated_length<<= 1;

	vbp->buffer = (char *)safe_realloc(vbp->buffer,vbp->allocated_length);
}

/*
 * Append buffer/length style buffer.
 */
int variable_buffer_append(struct varbuf *vbp,char *append,size_t length)
{
	if (!length)
		return vbp->buffer_length;

	variable_buffer_size_extend(vbp,vbp->buffer_length + length + 1);

	vbp->last_append_offset = vbp->buffer_length;

	memcpy(vbp->buffer + vbp->buffer_length,append,length);
	vbp->buffer_length+= length;
	vbp->buffer[vbp->buffer_length] = '\0';	/* Always ASCIIZ terminate */

	return vbp->buffer_length - vbp->header_skip_length;
}

/*
 * Append ASCIIZ string.
 */
int variable_buffer_append_string(struct varbuf *vbp,char *append)
{
	return variable_buffer_append(vbp,append,strlen(append));
}

/*
 * Append printf-format data.
 */
int variable_buffer_append_printf(struct varbuf *vbp,char *fmt,...)
{
	va_list args;
	char *p;
	int retval;

	va_start(args,fmt);
	vasprintf(&p,fmt,args);
	va_end(args);
	retval = variable_buffer_append_string(vbp,p);
	free(p);
	return retval;
}

/*
 * Append vprintf-format data.
 */
int variable_buffer_append_vprintf(struct varbuf *vbp,char *fmt,va_list args)
{
	char *p;
	int retval;

	vasprintf(&p,fmt,args);
	retval = variable_buffer_append_string(vbp,p);
	free(p);
	return retval;
}

/*
 * Skip (drop) specified number bytes at head of buffer.
 */
int variable_buffer_skip_head(struct varbuf *vbp,size_t length)
{
	if (length > (vbp->buffer_length - vbp->header_skip_length))
		length = vbp->buffer_length - vbp->header_skip_length;

	vbp->header_skip_length+= length;

	if (vbp->header_skip_length + length > vbp->buffer_length){
		/*
		 * Heurstic: when we have skipped more than half of the
		 * buffer, physically move and resize it to avoid wasting
		 * memory while still being somewhat efficient.
		 */
		memmove(vbp->buffer,
		        vbp->buffer + vbp->header_skip_length,
		        vbp->buffer_length - vbp->header_skip_length);
		vbp->buffer_length-= vbp->header_skip_length;
		vbp->header_skip_length = 0;
		/* realloc() here, maybe? */
		return vbp->buffer_length;	/* Clean and fresh */
	}

	return vbp->buffer_length - vbp->header_skip_length;
}

/*
 * Place raw buffer in pointer specified and return length.
 */
int variable_buffer_get(struct varbuf *vbp,char **buffer)
{
	if (buffer)
		*buffer = vbp->buffer + vbp->header_skip_length;
	return vbp->buffer_length - vbp->header_skip_length;
}

/*
 * Return buffer in ASCIIZ form.
 */
char *variable_buffer_get_string(struct varbuf *vbp)
{
	/*
	 * We always '\0' terminate when appending,
	 * so there is nothing special here.
	 */
	return vbp->buffer + vbp->header_skip_length;
}

/*
 * Detach (clear without freeing and return pointer for) buffer
 * in ASCIIZ form.
 */
char *variable_buffer_detach_string(struct varbuf *vbp)
{
	char *buffer;
	if (vbp->header_skip_length){
		/*
		 * We need to return a clean string.
		 */
		memmove(vbp->buffer,
		        vbp->buffer + vbp->header_skip_length,
		        vbp->buffer_length - vbp->header_skip_length);
		vbp->buffer_length-= vbp->header_skip_length;
		vbp->header_skip_length = 0;
		/* realloc() here, maybe? */
	}
	buffer = vbp->buffer;

	variable_buffer_init(vbp);
	return buffer;
}

/*
 * Return last appended string in ASCIIZ form.
 */
char *variable_buffer_get_last_appended_string(struct varbuf *vbp){
	return vbp->buffer + vbp->last_append_offset;
}

/*---------------------------------------------------------------------------*
 *
 * File functions.
 *
 *---------------------------------------------------------------------------*/

/*
 * read() with fill-all-of-buffer retries and error handling.
 */
ssize_t safe_read(int fd,unsigned char *buf,size_t len){
	int r,ofs = 0;

	for (;;){
		r = read(fd,buf + ofs,len - ofs);

		if (r > 0){
			ofs+= r;
			if (ofs == len)
				return ofs;
		} else if (r == 0){
			return ofs;
		} else {
			if ((errno == EAGAIN) ||
			    (errno == EINTR) ||
			    (errno == ERESTART))
				continue;
			return -1;
		}
	}
}

/*
 * write() with write-all-of-buffer retries and error handling.
 */
ssize_t safe_write(int fd,unsigned char *buf,size_t len){
	int r,ofs = 0;

	for (;;){
		r = write(fd,buf + ofs,len - ofs);

		if (r > 0){
			ofs+= r;
			if (ofs == len)
				return ofs;
		} else if (r == 0){
			return ofs;
		} else {
			if ((errno == EAGAIN) ||
			    (errno == EINTR) ||
			    (errno == ERESTART))
				continue;
			return -1;
		}
	}
}

/*
 * Lined file reader.  Calls given callback for each line.
 */

int file_read_fd(int fd,line_callback_t callback,void *data){
	char *s,*line;
	int remaining,buf_ofs,chunk_length,line_ofs;
	char buf[CHUNK_SIZE+1];

	line = NULL;
	line_ofs = 0;
	while ((remaining = safe_read(fd,buf,CHUNK_SIZE))){
		if (remaining == -1)
			return 0;
		for (buf_ofs = 0;remaining > 0;){
			s = memchr(buf + buf_ofs,'\n',remaining);
			if (s){
				chunk_length = s - buf - buf_ofs;
				if (chunk_length){	/* Not empty line/end of continued line */
					if (!line){
						/* FAST PATH: Unsplit line */
						buf[buf_ofs + chunk_length] = '\0';
						(*callback)(data,buf + buf_ofs);
						/* line stays NULL */
					} else {
						line = safe_realloc(line,line_ofs + chunk_length + 1);
						memcpy(line + line_ofs,buf + buf_ofs,chunk_length);
						line_ofs+= chunk_length;
					}
				}
				if (line){		/* Not empty line */
					line[line_ofs] = '\0';
					(*callback)(data,line);
					free(line);
					line = NULL;
					line_ofs = 0;
				}
				chunk_length++;	/* Count \n */
			} else {
				if (!line)
					line = safe_malloc(remaining + 1);
				else
					line = safe_realloc(line,line_ofs + remaining + 1);
				memcpy(line + line_ofs,buf + buf_ofs,remaining);
				chunk_length = remaining;
				line_ofs+= chunk_length;
			}
			remaining-= chunk_length;
			buf_ofs+= chunk_length;
		}
	}
	return 1;
}

int file_read(char *file,line_callback_t callback,void *data){
	int fd,r,serrno;

	fd = open(file,O_RDONLY);
	if (fd != -1){
		r = file_read_fd(fd,callback,data);
		serrno = errno;
		close(fd);
		errno = serrno;
		return r;
	}
	return 0;
}
